## preallocation

function _preallocate(::Type{T}, sizes::Tuple, sizehint) where {T}
    res = spzeros(T, sizes...)
    @static if VERSION >= v"1.7"
        sizehint!(res, first(sizes) * sizehint)
    end

    return res
end

__an_elt(χ::Character) = first(first(conjugacy_classes(χ)))
__an_elt(α::AlgebraElement) = first(basis(parent(α)))

_projection_size(p::PermutationGroups.AbstractPerm) = (d = degree(p); (d, d))
_projection_size(m::AbstractMatrix) = size(m)

_projection_size(::Nothing, χ::Character) = _projection_size(__an_elt(χ))
_projection_size(::Nothing, α::AlgebraElement) = _projection_size(__an_elt(α))
function _projection_size(hom::InducedActionHomomorphism, χ::Character)
    return _projection_size(induce(hom, __an_elt(χ)))
end
function _projection_size(hom::InducedActionHomomorphism, α::AlgebraElement)
    return _projection_size(induce(hom, __an_elt(α)))
end

_hint(χ::Character) = length(conjugacy_classes(χ))
_hint(α::AlgebraElement) = count(!iszero, StarAlgebras.coeffs(α))

function preallocate(::Type{T}, χ::Union{Character,AlgebraElement}) where {T}
    return preallocate(T, nothing, χ)
end

function preallocate(
    hom::InducedActionHomomorphism,
    χ::Union{Character,AlgebraElement},
)
    T = Base._return_type(*, Tuple{coeff_type(hom),eltype(χ)})
    return preallocate(T, hom, χ)
end

function preallocate(
    ::Type{T},
    a::Union{Nothing,InducedActionHomomorphism},
    χ::Union{Character,AlgebraElement},
) where {T}
    sizes = _projection_size(a, χ)
    return _preallocate(T, sizes, _hint(χ))
end

## matrix projection [irreducible]
"""
    matrix_projection([hom::InducedActionHomomorphism, ]χ::Character{T})
Compute matrix projection associated to character `χ`.

Returned `M<:AbstractMatrix{T}` of size `(d, d)` where the degree `d` of the
projecion is determined by elements in `conjugacy_classes(χ)`. E.g. `d` could
be equal to the `degree` when conjugacy classes consist of `AbstractPerms`.
If the homomorphism is passed, the dimension will be derived in similar
manner from the elements of the image of the homomorphism.

The precise type of `M` can be altered by overloading

```
preallocate(::Type{T}, [hom::InducedActionHomomorphism, ]χ::Character)
```
"""
function matrix_projection(χ::Character{T}) where {T}
    return _mproj_outsT!(preallocate(T, χ), χ)
end

function matrix_projection(
    hom::InducedActionHomomorphism,
    χ::Character{T},
) where {T}
    return _mproj_outsT!(preallocate(hom, χ), hom, χ)
end

function matrix_projection(χ::Character{T}) where {T<:Rational}
    if all(isone ∘ Cyclotomics.conductor, table(χ))
        return _mproj_fitsT!(preallocate(T, χ), χ)
    else
        return _mproj_outsT!(preallocate(T, χ), χ)
    end
end

function matrix_projection(χ::Character{T}) where {T<:Union{Cyclotomic,Complex}}
    return _mproj_fitsT!(preallocate(T, χ), χ)
end
function matrix_projection(
    hom::InducedActionHomomorphism,
    χ::Character{T},
) where {T<:Union{Cyclotomic,Complex}}
    return _mproj_fitsT!(preallocate(hom, χ), hom, χ)
end

function matrix_projection(χ::Character{T}) where {T<:AbstractFloat}
    if all(isreal, table(χ))
        return _mproj_fitsT!(preallocate(T, χ), χ)
    else
        return _mproj_outsT!(preallocate(T, χ), χ)
    end
end

function matrix_projection(
    hom::InducedActionHomomorphism,
    χ::Character{T},
) where {T<:AbstractFloat}
    if all(isreal, table(χ))
        return _mproj_fitsT!(preallocate(hom, χ), hom, χ)
    else
        return _mproj_outsT!(preallocate(hom, χ), hom, χ)
    end
end

_mproj_fitsT!(args...) = _mproj_outsT!(args...)

function _mproj_outsT!(mproj::AbstractMatrix{T}, χ::Character) where {T}
    mproj .= sum(
        c .* matrix_projection_irr(Character(table(χ), i)) for
        (i, c) in pairs(constituents(χ)) if !iszero(c)
    )
    return mproj
end

function _mproj_outsT!(
    mproj::AbstractMatrix{T},
    hom::InducedActionHomomorphism,
    χ::Character,
) where {T}
    mproj .= sum(
        c .* matrix_projection_irr(hom, Character(table(χ), i)) for
        (i, c) in pairs(constituents(χ)) if !iszero(c)
    )
    return mproj
end

"""
    matrix_projection_irr([::Type, ]χ::Character)
    matrix_projection_irr(hom::InducedActionHomomorphism, χ::Character)
Compute matrix projection associated to an *irreducible* character `χ`.

The returned matrix defines so called *isotypical* projection.

See also [matrix_projection](@ref).
"""
matrix_projection_irr(χ::Character{T}) where {T} = matrix_projection_irr(T, χ)
function matrix_projection_irr(::Type{T}, χ::Character) where {T}
    return matrix_projection_irr_acc!(preallocate(T, χ), χ, 1)
end

function matrix_projection_irr(hom::InducedActionHomomorphism, χ::Character)
    return matrix_projection_irr_acc!(preallocate(hom, χ), hom, χ, 1)
end

function matrix_projection_irr_acc!(
    result::AbstractMatrix,
    χ::Character,
    weight,
)
    @assert isirreducible(χ)
    LinearAlgebra.checksquare(result)
    ccls = conjugacy_classes(χ)
    vals = collect(values(χ))
    weight *= degree(χ) // sum(length, ccls)
    result = matrix_projection_irr_acc!(result, vals, ccls, weight)
    return result
end

function matrix_projection_irr_acc!(
    result::AbstractMatrix,
    vals,
    ccls::AbstractVector{<:AbstractOrbit{<:PermutationGroups.AbstractPerm}},
    weight,
)
    iszero(weight) && return result
    I = UInt32[]
    J = UInt32[]
    V = eltype(result)[]
    for (val, cc) in zip(vals, ccls)
        iszero(val) && continue
        w = weight * val
        for g in cc
            for i in 1:size(result, 1)
                push!(I, i)
                push!(J, i^g)
                push!(V, w)
            end
        end
    end
    result += sparse(I, J, V)
    return result
end

function matrix_projection_irr_acc!(
    res::AbstractMatrix,
    vals,
    ccls::AbstractVector{<:AbstractOrbit{<:AbstractMatrix}},
    weight,
)
    # TODO: call to inv(Matrix(g)) is a dirty hack, since if `g`
    # is given by a sparse matrix `inv(g)` will fail.
    for (val, cc) in zip(vals, ccls)
        iszero(val) && continue
        for g in cc
            res .+= (weight * val) .* inv(convert(Matrix, g))
        end
    end
    return res
end

function matrix_projection_irr_acc!(
    result::AbstractMatrix,
    hom::InducedActionHomomorphism,
    χ::Character,
    weight,
)
    @assert isirreducible(χ)
    LinearAlgebra.checksquare(result)
    vals = collect(values(χ))
    ccls = conjugacy_classes(χ)
    weight *= degree(χ) // sum(length, ccls)
    result = matrix_projection_irr_acc!(result, hom, vals, ccls, weight)
    return result
end

function matrix_projection_irr_acc!(
    result::AbstractMatrix,
    hom::InducedActionHomomorphism{<:ByPermutations},
    class_values,
    conjugacy_cls,
    weight,
)
    iszero(weight) && return result
    I = UInt32[]
    J = UInt32[]
    V = eltype(result)[]
    for (val, ccl) in zip(class_values, conjugacy_cls)
        iszero(val) && continue
        w = weight * val
        for g in ccl
            h = induce(hom, g)
            @assert h isa PermutationGroups.Perm
            for i in 1:size(result, 1)
                push!(I, i)
                push!(J, i^h)
                push!(V, w)
            end
        end
    end
    result += sparse(I, J, V)
    return result
end

function matrix_projection_irr_acc!(
    result::AbstractMatrix,
    hom::InducedActionHomomorphism{<:ByLinearTransformation},
    class_values,
    conjugacy_cls,
    weight,
)
    iszero(weight) && return result
    for (val, cc) in zip(class_values, conjugacy_cls)
        iszero(val) && continue
        w = weight * val
        for g in cc
            result .+= w .* induce(hom, inv(g))
        end
    end
    return result
end

"""
    matrix_representation([::Type{T}, ]α::AlgebraElement)
    matrix_representation([::Type{T}, ]hom::InducedActionHomomorphism, α::AlgebraElement)
Compute matrix representative of the given algebra element `α`.

If `α` is an element of the group (or monoid) algebra `RG` of `G`, then every
linear representation `π : G → GL(V)` gives rise to an algebra homomorphism
`π : RG → B(V)` (bounded operators on `V`). This function evaluates this
representation given either by
 * the natural action of `G` on some `V` (e.g. every permutation group of degree
 `d` acts on `d`-dimensional `V` by basis vector permutation), or
 * by action homomorphism `hom`.

See also [matrix_projection](@ref).
"""
matrix_representation(α::AlgebraElement) = matrix_representation(eltype(α), α)
function matrix_representation(::Type{T}, α::AlgebraElement) where {T}
    return matrix_representation_acc!(preallocate(T, α), α)
end

function matrix_representation(
    hom::InducedActionHomomorphism,
    α::AlgebraElement,
)
    return matrix_representation_acc!(preallocate(hom, α), hom, α)
end

function matrix_representation_acc!(
    result::AbstractMatrix,
    α::AlgebraElement{
        <:StarAlgebra{<:PermutationGroups.AbstractPermutationGroup},
    },
)
    b = basis(parent(α))
    I = UInt32[]
    J = UInt32[]
    V = eltype(result)[]
    for (idx, val) in StarAlgebras._nzpairs(StarAlgebras.coeffs(α))
        g = b[idx]
        iszero(val) && continue
        for i in 1:size(result, 1)
            push!(I, i)
            push!(J, i^g)
            push!(V, val)
        end
    end
    result += sparse(I, J, V)
    return result
end

function matrix_representation_acc!(
    result::AbstractMatrix,
    hom::InducedActionHomomorphism{<:ByPermutations},
    α::AlgebraElement,
)
    b = basis(parent(α))
    I = UInt32[]
    J = UInt32[]
    V = eltype(result)[]
    for (idx, val) in StarAlgebras._nzpairs(StarAlgebras.coeffs(α))
        iszero(val) && continue
        g = induce(hom, b[idx])
        @assert g isa PermutationGroups.Perm
        for i in 1:size(result, 1)
            push!(I, i)
            push!(J, i^g)
            push!(V, val)
        end
    end
    result += sparse(I, J, V)
    return result
end

function matrix_representation_acc!(
    result::AbstractMatrix,
    hom::InducedActionHomomorphism{<:ByLinearTransformation},
    α::AlgebraElement,
)
    b = basis(parent(α))
    for (idx, val) in StarAlgebras._nzpairs(StarAlgebras.coeffs(α))
        iszero(val) && continue
        result .+= val .* induce(hom, inv(b[idx]))
    end

    return result
end

# convenience for uniform calls below
function matrix_representation(χ::Character)
    return isirreducible(χ) ? matrix_projection_irr(χ) : matrix_projection(χ)
end

function matrix_representation(hom::InducedActionHomomorphism, χ::Character)
    return isirreducible(χ) ? matrix_projection_irr(hom, χ) :
           matrix_projection(hom, χ)
end

## Finding basis of the row-space (right image) of an AbstractMatrix

"""
    image_basis(A::AbstractMatrix)
    image_basis([hom::InducedActionHomomorphism, ]χ::Character[, rank])
    image_basis([hom::InducedActionHomomorphism, ]α::AlgebraElement[, rank])
Return basis of the row-space of a matrix.

For characters or algebra elements return a basis of the row-space of
`matrix_projection([hom, ]χ)` or `matrix_representation([hom, ]α)`.

Two methods are employed to achieve the goal.
* By default (symbolic, exact) row echelon form is produced, and therefore there
is no guarantee on the orthogonality of the returned basis vectors (rows).
* If `eltype(A) <: AbstractFloat` a thin `svd` (or `qr` in the sparse case)
decomposition is computed and the appropriate rows of its orthogonal factor are
returned, thus the basis is (numerically) orthonormal.

If the dimension (`=rank` of the matrix) is known beforehand (because of e.g.
algebra) it can be passed to `image_basis` to use more efficient methods.

# Examples:
```julia
julia> a = rand(-3:3, 3,3).//rand(1:4, 3,3);

julia> a[3, :] .= 2a[1, :] .- 1a[2, :]; a
3×3 Matrix{Rational{Int64}}:
  3//2   0//1  1//1
 -1//2  -2//1  1//2
  7//2   2//1  3//2

julia> ib = SymbolicWedderburn.image_basis(a)
2×3 Matrix{Rational{Int64}}:
 1//1  0//1   2//3
 0//1  1//1  -5//12

julia> ibf = SymbolicWedderburn.image_basis(float.(a))
2×3 Matrix{Float64}:
 -0.666651  0.416657  -0.618041
  0.529999  0.847998  -8.85356e-17

```
"""
image_basis(A::AbstractMatrix, rank = nothing) = image_basis!(deepcopy(A), rank)

function image_basis(α::Union{Character,AlgebraElement}, rank = nothing)
    mpr = matrix_representation(α)
    return image_basis!(mpr, rank)
end

function image_basis(
    hom::InducedActionHomomorphism,
    α::Union{Character,AlgebraElement},
    rank = nothing,
)
    mpr = matrix_representation(hom, α)
    return image_basis!(mpr, rank)
end

##
# the internal, (possibly) modifying versions

_eps(T::Type{<:AbstractFloat}) = eps(T)
_eps(::Type{Complex{T}}) where {T} = 2 * _eps(real(T))
_eps(::Type{<:Cyclotomic{T}}) where {T} = 2_eps(T)
_eps(::Any) = 0.0

function image_basis!(A::AbstractMatrix, rank = nothing)
    img, pivots = row_echelon_form!(A)
    if img isa AbstractSparseArray
        ε = _eps(eltype(img))
        if iszero(ε)
            dropzeros!(img)
        else
            droptol!(img, _eps(eltype(img)) * max(size(img)...))
        end
    end
    # TODO: orthogonalize the result
    return img[pivots, :]
end

function image_basis!(
    A::AbstractMatrix{T},
    ::Nothing,
) where {T<:Union{AbstractFloat,Complex}}
    F = svd!(A)
    tol = _eps(T) * first(F.S)
    rk = count(x -> x > tol, F.S)
    return Matrix((@view F.U[1:rk, :])')
end

function image_basis!(
    A::AbstractSparseMatrix{T},
    ::Nothing,
) where {T<:Union{AbstractFloat,Complex}}
    F = qr(A)
    rk = rank(F)
    img = let tmp = F.Q * Matrix(I, size(A, 1), rk)
        pinv = getfield(F, :rpivinv)
        sparse((@view tmp[pinv, :])')
    end
    return droptol!(img, _eps(T) * max(size(img)...))
end

# to disambiguate
function image_basis!(
    M::AbstractMatrix{T},
    rank::Integer,
) where {T<:Union{AbstractFloat,Complex}}
    img = image_basis!(M, nothing)
    if size(img, 1) ≠ rank
        @warn "Possibly wrong numerical rank: (numerical) $(size(img, 1)) ≠ $rank (expected)"
    end
    return img
end

function image_basis!(
    M::AbstractSparseMatrix{T},
    rank::Integer,
) where {T<:Union{AbstractFloat,Complex}}
    # return _image_basis!(M, rank)
    img = image_basis!(M, nothing)
    if size(img, 1) ≠ rank
        @warn "Possibly wrong rank estimate? (numerical) $(size(img, 1)) ≠ $rank (expected)"
    end
    return img
end
